=== modified file 'bzrlib/delta.py'
--- bzrlib/delta.py	2007-03-07 23:15:10 +0000
+++ bzrlib/delta.py	2007-04-16 01:00:20 +0000
@@ -195,6 +195,12 @@
             print >>to_file, 'unknown:'
             show_list(self.unversioned)
 
+    def get_changes_as_text(self, show_ids=False, show_unchanged=False,
+             short_status=False):
+        import StringIO
+        output = StringIO.StringIO()
+        self.show(output, show_ids, show_unchanged, short_status)
+        return output.getvalue()
 
 @deprecated_function(zero_nine)
 def compare_trees(old_tree, new_tree, want_unchanged=False,

=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2007-03-15 23:24:59 +0000
+++ bzrlib/errors.py	2007-04-16 01:00:20 +0000
@@ -1521,7 +1521,19 @@
     def __init__(self, from_path, to_path, extra=None):
         BzrMoveFailedError.__init__(self, from_path, to_path, extra)
 
-
+class BzrRemoveChangedFilesError(BzrError):
+    """Used when user is trying to remove changed files."""
+
+    _fmt = ("Can't remove changed or unknown files:\n%(changes_as_text)s"
+        "Use --keep to not delete them, or --force to delete them regardless.")
+
+    def __init__(self, tree_delta):
+        BzrError.__init__(self)
+        self.changes_as_text = tree_delta.get_changes_as_text()
+        #self.paths_as_string = '\n'.join(changed_files)
+        #self.paths_as_string = '\n'.join([quotefn(p) for p in changed_files])
+        
+    
 class BzrBadParameterNotString(BzrBadParameter):
 
     _fmt = "Parameter %(param)s is not a string or unicode string."

=== modified file 'bzrlib/tests/workingtree_implementations/test_remove.py'
--- bzrlib/tests/workingtree_implementations/test_remove.py	2007-04-14 12:40:37 +0000
+++ bzrlib/tests/workingtree_implementations/test_remove.py	2007-04-16 01:00:20 +0000
@@ -16,12 +16,15 @@
 
 """Tests for interface conformance of 'WorkingTree.remove'"""
 
+import re
 from bzrlib.tests.workingtree_implementations import TestCaseWithWorkingTree
+from bzrlib import errors, osutils
 
 class TestRemove(TestCaseWithWorkingTree):
     """Tests WorkingTree.remove"""
 
     files=['a', 'b/', 'b/c']
+    a = ['a']
     b = ['b']
     b_c = ['b', 'b/c']
 
@@ -31,12 +34,21 @@
         self.build_tree(TestRemove.files)
         return tree
 
+    def test_remove_keep(self):
+        """Check that files are unversioned but not deleted."""
+        tree = self.getTree()
+        tree.add(TestRemove.files)
+        self.assertInWorkingTree(TestRemove.files)
+
+        tree.remove(TestRemove.files)
+        self.assertNotInWorkingTree(TestRemove.files)
+        self.failUnlessExists(TestRemove.files)
+
     def test_remove_unchanged_files(self):
-        """check that unchanged files are removed and deleted."""
+        """Check that unchanged files are removed and deleted."""
         tree = self.getTree()
         tree.add(TestRemove.files)
         tree.commit("files must not have changes")
-
         self.assertInWorkingTree(TestRemove.files)
 
         tree.remove(TestRemove.files, keep_files=False)
@@ -44,19 +56,57 @@
         self.assertNotInWorkingTree(TestRemove.files)
         self.failIfExists(TestRemove.files)
 
-    def test_remove_changed_files(self):
-        """check that changed files are removed but not deleted."""
+    def test_remove_added_files(self):
+        """Removal of newly added files must fail."""
         tree = self.getTree()
         tree.add(TestRemove.files)
         self.assertInWorkingTree(TestRemove.files)
-
-        tree.remove(TestRemove.files, keep_files=False)
-
-        self.assertNotInWorkingTree(TestRemove.files)
+        try:
+            tree.remove(TestRemove.files, keep_files=False)
+            self.fail('Should throw BzrRemoveChangedFilesError')
+        except errors.BzrRemoveChangedFilesError, e:
+            self.assertTrue(re.match('Can\'t remove changed or unknown files:'
+                '.*added:.*a.*b.*b/c.*',
+                str(e), re.DOTALL))
+        self.assertInWorkingTree(TestRemove.files)
         self.failUnlessExists(TestRemove.files)
 
+    def test_remove_changed_file(self):
+        """Removal of a changed files must fail."""
+        tree = self.getTree()
+        tree.add(TestRemove.a)
+        tree.commit("make sure a is versioned")
+        f = file('a', 'wb')
+        f.write("some other new content!")
+        f.close()
+        self.assertInWorkingTree(TestRemove.a)
+        try:
+            tree.remove(TestRemove.a, keep_files=False)
+            self.fail('Should throw BzrRemoveChangedFilesError')
+        except errors.BzrRemoveChangedFilesError, e:
+            self.assertTrue(re.match('Can\'t remove changed or unknown files:'
+                '.*modified:.*a.*',
+                str(e), re.DOTALL))
+        self.assertInWorkingTree(TestRemove.a)
+        self.failUnlessExists(TestRemove.a)
+
+    def test_remove_deleted_files(self):
+        """Check that files are removed if they don't exist any more."""
+        tree = self.getTree()
+        tree.add(TestRemove.files)
+        tree.commit("make sure files are versioned")
+        for f in ['b/c','b','a']:
+            osutils.delete_any(f)
+        self.assertInWorkingTree(TestRemove.files)
+        self.failIfExists(TestRemove.files)
+
+        tree.remove(TestRemove.files, keep_files=False)
+
+        self.assertNotInWorkingTree(TestRemove.files)
+        self.failIfExists(TestRemove.files)
+
     def test_force_remove_changed_files(self):
-        """check that changed files are removed and deleted when forced."""
+        """Check that changed files are removed and deleted when forced."""
         tree = self.getTree()
         tree.add(TestRemove.files)
         self.assertInWorkingTree(TestRemove.files)
@@ -66,14 +116,31 @@
         self.assertNotInWorkingTree(TestRemove.files)
         self.failIfExists(TestRemove.files)
 
+    def test_remove_unknown_files(self):
+        """Try to delete unknown files."""
+        tree = self.getTree()
+        try:
+            tree.remove(TestRemove.files, keep_files=False)
+            self.fail('Should throw BzrRemoveChangedFilesError')
+        except errors.BzrRemoveChangedFilesError, e:
+            self.assertTrue(re.match('Can\'t remove changed or unknown files:'
+                '.*unknown:.*b/c.*b.*a.*',
+                str(e), re.DOTALL))
+
     def test_remove_nonexisting_files(self):
-        """delete files which does not exist."""
+        """Try to delete non-existing files."""
         tree = self.getTree()
-        tree.remove(TestRemove.files, keep_files=False)
         tree.remove([''], keep_files=False)
-        tree.remove(TestRemove.b, keep_files=False)
+        try:
+            tree.remove(['xyz', 'abc/def'], keep_files=False)
+            self.fail('Should throw BzrRemoveChangedFilesError')
+        except errors.BzrRemoveChangedFilesError, e:
+            self.assertTrue(re.match('Can\'t remove changed or unknown files:'
+                '.*unknown:.*xyz.*abc/def.*',
+                str(e), re.DOTALL))
 
     def test_remove_nonempty_directory(self):
+        """Unchanged non-empty directories should be deleted."""
         tree = self.getTree()
         tree.add(TestRemove.files)
         tree.commit("make sure b is versioned")
@@ -81,6 +148,26 @@
         self.failUnlessExists(TestRemove.files)
         tree.remove(TestRemove.b, keep_files=False)
         self.assertNotInWorkingTree(TestRemove.b)
+        self.failIfExists(TestRemove.b)
+        
+    def test_remove_nonempty_directory_with_unknowns(self):
+        """Unchanged non-empty directories should be deleted."""
+        tree = self.getTree()
+        tree.add(TestRemove.files)
+        tree.commit("make sure b is versioned")
+        self.assertInWorkingTree(TestRemove.files)
+        self.failUnlessExists(TestRemove.files)
+        f = file('b/my_unknown_file', 'wb')
+        f.write("some content!")
+        f.close()
+        try:
+            tree.remove(TestRemove.b, keep_files=False)
+            self.fail('Should throw BzrRemoveChangedFilesError')
+        except errors.BzrRemoveChangedFilesError, e:
+            self.assertTrue(re.match('Can\'t remove changed or unknown files:'
+                '.*unknown:.*b/my_unknown_file.*',
+                str(e), re.DOTALL))
+        self.assertInWorkingTree(TestRemove.b)
         self.failUnlessExists(TestRemove.b)
 
     def test_force_remove_nonempty_directory(self):
@@ -93,12 +180,26 @@
         self.assertNotInWorkingTree(TestRemove.b_c)
         self.failIfExists(TestRemove.b_c)
 
-    def test_remove_keep(self):
-        """check that files are unversioned but not delete."""
+    def test_remove_directory_with_changed_file(self):
+        """Refuse to delete directories with changed files."""
         tree = self.getTree()
-        tree.add(TestRemove.files)
-        self.assertInWorkingTree(TestRemove.files)
+        tree.add(TestRemove.b_c)
+        tree.commit("make sure b and c are versioned")
+        f = file('b/c', 'wb')
+        f.write("some other new content!")
+        f.close()
+        self.assertInWorkingTree(TestRemove.b_c)
+        try:
+            tree.remove(TestRemove.b, keep_files=False)
+            self.fail('Should throw BzrRemoveChangedFilesError')
+        except errors.BzrRemoveChangedFilesError, e:
+            self.assertTrue(re.match('Can\'t remove changed or unknown files:'
+                '.*modified:.*b/c.*',
+                str(e), re.DOTALL))
+        self.assertInWorkingTree(TestRemove.b_c)
+        self.failUnlessExists(TestRemove.b_c)
 
-        tree.remove(TestRemove.files)
-        self.assertNotInWorkingTree(TestRemove.files)
-        self.failUnlessExists(TestRemove.files)
+        #see if we can force it now..
+        tree.remove(TestRemove.b, keep_files=False, force=True)
+        self.assertNotInWorkingTree(TestRemove.b_c)
+        self.failIfExists(TestRemove.b_c)

=== modified file 'bzrlib/workingtree.py'
--- bzrlib/workingtree.py	2007-04-14 12:40:37 +0000
+++ bzrlib/workingtree.py	2007-04-16 01:00:20 +0000
@@ -363,6 +363,10 @@
     def abspath(self, filename):
         return pathjoin(self.basedir, filename)
     
+    def canonicalpath(self, filename):
+        """Normanize filename"""
+        return self.relpath(self.abspath(filename))
+    
     def basis_tree(self):
         """Return RevisionTree for the current last revision.
         
@@ -1779,22 +1783,45 @@
             files = [files]
 
         inv = self.inventory
-        changed_files=[]
-
-        # Get file names into canonical form.
-        files = [self.relpath(self.abspath(filename))
-            for filename in files]
+
+        new_files=set()
+        unknown_files_in_directory=set()
+
+        def recurse_directory_to_add_files(directory):
+            # recurse directory and add all files 
+            # so we can check if they have changed.
+            for contained_dir_info in self.walkdirs(directory):
+                for file_info in contained_dir_info[1]:
+                    if file_info[2] == 'file':
+                        canonicalpath = self.canonicalpath(file_info[0])
+                        if file_info[4]: #is it versioned?
+                            new_files.add(canonicalpath)
+                        else:
+                            unknown_files_in_directory.add(
+                                (canonicalpath, None, file_info[2]))
+
+        for filename in files:
+            # Get file name into canonical form.
+            filename = self.canonicalpath(filename)
+            if len(filename) > 0:
+                new_files.add(filename)
+                if osutils.isdir(filename):
+                    recurse_directory_to_add_files(filename)
+        files = [f for f in new_files]
 
         # Sort needed to first handle directory content before the directory
         files.sort(reverse=True)
-
         if not keep_files and not force:
-            changes = self.changes_from(self.basis_tree(),
+            tree_delta = self.changes_from(self.basis_tree(),
                 specific_files=files)
-            for f in changes.modified:
-                changed_files.append(f[0])
-            for f in changes.added:
-                changed_files.append(f[0])
+            for unknown_file in unknown_files_in_directory:
+                tree_delta.unversioned.extend((unknown_file,))
+            if bool(tree_delta.modified
+                    or tree_delta.added
+                    or tree_delta.renamed
+                    or tree_delta.kind_changed
+                    or tree_delta.unversioned):
+                raise errors.BzrRemoveChangedFilesError(tree_delta)
 
         # do this before any modifications
         for f in files:
@@ -1817,28 +1844,16 @@
 
             if not keep_files:
                 if osutils.lexists(f):
-                    if force:
-                        # recursively delete f
-                        if osutils.isdir(f):
-                            osutils.rmtree(f)
-                        else:
-                            os.unlink(f)
+                    if osutils.isdir(f) and len(os.listdir(f)) > 0:
+                        message="%s is not empty directory "\
+                            "and won't be deleted." % (f,)
+                    else:
+                        osutils.delete_any(f)
                         message="deleted %s" % (f,)
-                    elif fid: 
-                        # only consider deleting versioned files
-                        if f in changed_files:
-                            message="%s has changed and won't be deleted."\
-                                % (f,)
-                        elif osutils.isdir(f) and len(os.listdir(f)) > 0:
-                            message="%s is not empty directory "\
-                                "and won't be deleted." % (f,)
-                        else:
-                            osutils.delete_any(f)
-                            message="deleted %s" % (f,)
                 elif message is not None:
                     # only care if we haven't done anything yet.
                     message="%s does not exist." % (f,)
-                
+
             # print only one message (if any) per file.
             if message is not None:
                 note(message)
@@ -2144,11 +2159,16 @@
     def walkdirs(self, prefix=""):
         """Walk the directories of this tree.
 
+        returns a generator which yields items in the form:
+                ((curren_directory_path, fileid), 
+                 [(file1_path, file1_name, file1_kind, (lstat), file1_id,
+                   file1_kind), ... ])
+
         This API returns a generator, which is only valid during the current
         tree transaction - within a single lock_read or lock_write duration.
 
-        If the tree is not locked, it may cause an error to be raised, depending
-        on the tree implementation.
+        If the tree is not locked, it may cause an error to be raised,
+        depending on the tree implementation.
         """
         disk_top = self.abspath(prefix)
         if disk_top.endswith('/'):
@@ -2245,6 +2265,14 @@
                     disk_finished = True
 
     def _walkdirs(self, prefix=""):
+        """Walk the directories of this tree.
+
+           :prefix: is used as the directrory to start with.
+           returns a generator which yields items in the form:
+                ((curren_directory_path, fileid), 
+                 [(file1_path, file1_name, file1_kind, None, file1_id,
+                   file1_kind), ... ])
+        """
         _directory = 'directory'
         # get the root in the inventory
         inv = self.inventory

